# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
import textwrap
from argparse import Namespace
from typing import Any, Dict, List

import aiolirest
from aioli import cli
from aioli.cli import errors, render
from aioli.common import api
from aioli.common.api import authentication
from aioli.common.api.errors import NotFoundException
from aioli.common.declarative_argparse import Arg, ArgsDescription, Cmd, Group
from aioli.common.util import (
    construct_arguments,
    construct_environment,
    launch_dashboard,
)
from aiolirest.models.autoscaling import Autoscaling
from aiolirest.models.deployment import Deployment, DeploymentState
from aiolirest.models.deployment_request import DeploymentRequest
from aiolirest.models.event_info import EventInfo
from aiolirest.models.security import Security


@authentication.required
def dashboard(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.DeploymentsApi(session)

    deployment: Deployment = lookup_deployment(args.name, api_instance)

    assert deployment.id is not None
    observability = api_instance.deployments_id_observability_get(deployment.id)
    launch_dashboard(args, observability.dashboard_url)


@authentication.required
def show_deployment(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.DeploymentsApi(session)

    deployment: Deployment = lookup_deployment(args.name, api_instance)

    # For a more useful display, replace the model ID with its name
    packaged_models_api = aiolirest.PackagedModelsApi(session)
    deployment.model = packaged_models_api.models_id_get(deployment.model).name

    d = deployment.to_dict()
    # Remove clusterName for now - INF-243
    d.pop("clusterName")
    d.pop("project")  # something for INF-1516

    # Get the model version for the deployment
    deployment_model_versions = packaged_models_api.models_versions_get(deployment.model)
    for r in deployment_model_versions:
        if r.deployment_id == d["id"]:
            d["version"] = r.mdl_version
    if args.json:
        render.print_json(d)
    else:
        print(render.format_object_as_yaml(d))


@authentication.required
def list_deployments(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.DeploymentsApi(session)
        response = api_instance.deployments_get()
    model_api = aiolirest.PackagedModelsApi(session)

    if args.json:
        render.print_json(format_json(response, model_api))
    elif args.yaml:
        print(render.format_object_as_yaml(format_json(response, model_api)))
    else:
        format_deployments(response, args, model_api)


def format_json(
    response: List[Deployment], model_api: aiolirest.PackagedModelsApi
) -> List[Dict[str, str]]:
    deps = []
    for d in response:
        # Don't use the d.to_json() method as it adds backslash escapes for double quote
        m_dict = d.to_dict()
        m_dict.pop("modifiedAt")
        m_dict.pop("project")  # something for INF-1516
        # Use model name instead of id
        model = model_api.models_id_get(d.model)
        deployment_model_versions = model_api.models_versions_get(model.name)
        for r in deployment_model_versions:
            if r.deployment_id == m_dict["id"]:
                m_dict["version"] = r.mdl_version
        m_dict.pop("id")
        m_dict["model"] = model.name
        m_dict.pop("clusterName", None)
        deps.append(m_dict)

    return deps


def format_deployments(
    response: List[Deployment],
    args: Namespace,
    packaged_models_api: aiolirest.PackagedModelsApi,
) -> None:
    def format_deployment(e: Deployment, models_api: aiolirest.PackagedModelsApi) -> List[Any]:
        model = models_api.models_id_get(e.model)
        state = e.state
        total_failures = 0

        if state is None:
            state = DeploymentState()
        else:
            failure_info = state.failure_info
            if failure_info:
                total_failures = len(failure_info)

        secondary_state = e.secondary_state
        if secondary_state is None:
            secondary_state = DeploymentState()

        assert e.security is not None

        auto_scaling = e.auto_scaling
        if auto_scaling is None:
            auto_scaling = Autoscaling()

        if total_failures > 1:
            e.status = f"{e.status}\n({total_failures} errors)"
        elif total_failures == 1:
            e.status = f"{e.status}\n({total_failures} error)"

        deployment_model_versions = models_api.models_versions_get(model.name)
        version = next(r.mdl_version for r in deployment_model_versions if r.deployment_id == e.id)
        result = [
            e.name,
            model.name,
            version,
            e.namespace,
            e.status,
            e.security.authentication_required,
            state.status,
            state.traffic_percentage,
        ]

        return result

    headers = [
        "Name",
        "Model",
        "Version",
        "Namespace",
        "Status",
        "Auth Required",
        "State",
        "Traffic %",
    ]

    values = [format_deployment(r, packaged_models_api) for r in response]
    render.tabulate_or_csv(headers, values, args.csv)


@authentication.required
def create(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.DeploymentsApi(session)

        sec = Security(authenticationRequired=False)
        if args.authentication_required is not None:
            sec.authentication_required = str2bool(args.authentication_required)

        auto = Autoscaling(
            metric=args.autoscaling_metric,
        )

        if args.autoscaling_target is not None:
            auto.target = args.autoscaling_target

        if args.autoscaling_max_replicas is not None:
            auto.max_replicas = args.autoscaling_max_replicas

        if args.autoscaling_min_replicas is not None:
            auto.min_replicas = args.autoscaling_min_replicas

        r = DeploymentRequest(
            name=args.name,
            model=args.model,
            security=sec,
            namespace=args.namespace,
            autoScaling=auto,
            canaryTrafficPercent=args.canary_traffic_percent,
            environment=construct_environment(args),
            arguments=construct_arguments(args),
            nodeSelectors=construct_node_selectors(args),
            priorityClassName=args.priority_class_name,
        )
        api_instance.deployments_post(r)


def lookup_deployment(name: str, api: aiolirest.DeploymentsApi) -> Deployment:
    for r in api.deployments_get():
        if r.name == name:
            return r
    raise NotFoundException(f"deployment {name} not found")


@authentication.required
def update(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.DeploymentsApi(session)
        found = lookup_deployment(args.deploymentname, api_instance)
        request = DeploymentRequest(
            name=found.name,
            namespace=found.namespace,
            security=found.security,
            model=found.model,
            autoScaling=found.auto_scaling,
            canaryTrafficPercent=found.canary_traffic_percent,
            goalStatus=found.goal_status,
            environment=found.environment,
            arguments=found.arguments,
            nodeSelectors=found.node_selectors,
            priorityClassName=found.priority_class_name,
        )

        if request.auto_scaling is None:
            # Not likely, but testing these prevents complaints from mypy
            raise api.errors.BadResponseException("Unexpected null result")

        if args.pause and args.resume:
            raise errors.CliError("--pause and --resume cannot be specified at the same time")

        if args.pause:
            request.goal_status = "Paused"

        if args.resume:
            request.goal_status = "Ready"

        if args.name is not None:
            request.name = args.name

        if args.model is not None:
            request.model = args.model

        if args.namespace is not None:
            request.namespace = args.namespace

        if args.autoscaling_min_replicas is not None:
            request.auto_scaling.min_replicas = args.autoscaling_min_replicas

        if args.autoscaling_max_replicas is not None:
            request.auto_scaling.max_replicas = args.autoscaling_max_replicas

        if args.autoscaling_metric is not None:
            request.auto_scaling.metric = args.autoscaling_metric

        if args.autoscaling_target is not None:
            request.auto_scaling.target = args.autoscaling_target

        if args.canary_traffic_percent is not None:
            request.canary_traffic_percent = args.canary_traffic_percent

        assert request.security is not None

        if args.authentication_required is not None:
            request.security.authentication_required = str2bool(args.authentication_required)

        if args.env is not None:
            request.environment = construct_environment(args)

        if args.arg is not None:
            request.arguments = construct_arguments(args)

        if args.node_selector is not None:
            request.node_selectors = construct_node_selectors(args)

        if args.priority_class_name is not None:
            request.priority_class_name = args.priority_class_name

        headers = {"Content-Type": "application/json"}
        assert found.id is not None
        api_instance.deployments_id_put(found.id, request, _headers=headers)


@authentication.required
def delete_deployment(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.DeploymentsApi(session)
        found = lookup_deployment(args.name, api_instance)
        assert found.id is not None
        api_instance.deployments_id_delete(found.id)


@authentication.required
def get_deployment_events(args: Namespace) -> None:
    def format_events(event: EventInfo) -> List[Any]:
        if args.csv:
            message = event.message
        else:
            if event.message is not None:
                message = "\n".join(textwrap.wrap(event.message, width=70))
            else:
                message = ""
        result = [
            event.time,
            event.reason,
            message,
            event.event_type,
        ]
        return result

    with cli.setup_session(args) as session:
        api_instance = aiolirest.DeploymentsApi(session)
        found = lookup_deployment(args.name, api_instance)
        assert found.id is not None
        events = api_instance.deployments_id_events_get(found.id)
        headers = [
            "Time",
            "Reason",
            "Message",
            "Event Type",
        ]
        values = [format_events(r) for r in events]
        if args.json:
            render.print_json([e.to_dict() for e in events])
        elif args.yaml:
            print(render.format_object_as_yaml([e.to_dict() for e in events]))
        else:
            render.tabulate_or_csv(headers, values, args.csv)


def str2bool(v: str) -> bool:
    if isinstance(v, bool):
        return v
    if v.lower() in ("true", "1"):
        return True
    elif v.lower() in ("false", "0"):
        return False
    else:
        raise errors.CliError(
            f"--authentication-required: invalid boolean (true/false) value '{v}'"
        )


def construct_node_selectors(args: Namespace) -> Dict[str, str]:
    node_selectors: Dict[str, str] = {}
    if args.node_selector is None:
        return node_selectors

    for entry in args.node_selector:
        if entry is None:
            continue
        # split to label & value
        the_split = entry.split("=", maxsplit=1)
        label: str = the_split[0]
        value: str = ""
        if len(the_split) > 1:
            value = the_split[1]
        node_selectors[label] = value
    return node_selectors


common_deployment_args: ArgsDescription = [
    Arg(
        "--authentication-required",
        nargs="?",
        const=True,
        help="Deployed model requires callers to provide authentication. "
        "Specify boolean value 'true' or 'false'. Value '1' or '0' can also "
        "be used for 'true' or 'false' value. "
        "When true, all interactions with the deployed service are required to be authenticated.",
    ),
    Arg("--namespace", help="The Kubernetes namespace to be used for the deployment"),
    Arg("--autoscaling-min-replicas", help="Minimum number of replicas", type=int),
    Arg(
        "--autoscaling-max-replicas",
        help="Maximum number of replicas created based upon demand",
        type=int,
    ),
    Arg("--autoscaling-metric", help="Metric name which controls autoscaling"),
    Arg("--autoscaling-target", help="Metric target value", type=int),
    Arg(
        "--canary-traffic-percent",
        help="Percent traffic to pass to new model version",
        type=int,
    ),
    Arg(
        "-a",
        "--arg",
        nargs="?",
        help="Argument to be added to the service command line. "
        "If specifying an argument that starts with a '-', use the form --arg=<your-argument>. "
        "Specifying any --arg replaces prior args with the arguments on this invocation. "
        "Use a single --arg with no value to clear all arguments.",
        action="append",
    ),
    Arg(
        "-e",
        "--env",
        nargs="?",
        help="Specifies an environment variable & value as name=value, "
        "to be passed to the launched container. "
        "Specifying any --env replaces prior environment vars with those on this invocation. "
        "Use a single --env with no value to clear all environment vars.",
        action="append",
    ),
    Arg(
        "--node-selector",
        nargs="?",
        help="Specifies a node selector label & value as label=value, "
        "to be passed to the launched container. Example kubernetes.io/arch=amd64. "
        "Specifying any --node-selector replaces prior selectors with those on this invocation. "
        "Use a single --node-selector with no value to clear all selectors.",
        action="append",
    ),
    Arg(
        "--priority-class-name",
        help="Priority Class Name to be used for prioritization of deployments",
    ),
]

main_cmd = Cmd(
    "d|eployment|s",
    None,
    "manage trained deployments",
    [
        # Inspection commands.
        Cmd(
            "list ls",
            list_deployments,
            "list deployments",
            [
                Arg("--csv", action="store_true", help="print as CSV"),
                Arg("--json", action="store_true", help="print as JSON"),
                Arg("--yaml", action="store_true", help="print as YAML"),
            ],
            is_default=True,
        ),
        # Create command.
        Cmd(
            "create",
            create,
            "create a deployment",
            [
                Arg(
                    "name",
                    help="The name of the deployment. Must begin with a letter, but may contain "
                    "letters, numbers, and hyphen",
                ),
                Arg(
                    "--model",
                    help=(
                        "The package model id, name or versioned-name (evaluated in that "
                        "order) to be deployed"
                    ),
                    required="true",
                ),
            ]
            + common_deployment_args,
        ),
        # dashboard command.
        Cmd(
            "dashboard",
            dashboard,
            "launch the deployment dashboard",
            [
                Arg(
                    "name",
                    help="The name of the deployment.",
                ),
            ],
        ),
        # Show command.
        Cmd(
            "show",
            show_deployment,
            "show a deployment",
            [
                Arg(
                    "name",
                    help="The name of the deployment.",
                ),
                Group(
                    Arg("--yaml", action="store_true", help="print as YAML", default=True),
                    Arg("--json", action="store_true", help="print as JSON"),
                ),
            ],
        ),
        # Update command
        Cmd(
            "update",
            update,
            "modify a deployment",
            [
                Arg("deploymentname", help="The name of the deployment"),
                Arg(
                    "--name",
                    help="The new name of the deployment. Must begin with a letter, but may "
                    "contain letters, numbers, and hyphen",
                ),
                Arg(
                    "--model",
                    help=(
                        "The package model id, name or versioned-name (evaluated in that "
                        "order) to be deployed"
                    ),
                ),
                Arg("--pause", action="store_true", help="Pause the deployment"),
                Arg("--resume", action="store_true", help="Resume the deployment"),
            ]
            + common_deployment_args,
        ),
        Cmd(
            "delete",
            delete_deployment,
            "delete a deployment",
            [
                Arg("name", help="The name of the deployment"),
            ],
        ),
        Cmd(
            "event|s",
            get_deployment_events,
            "get deployment events",
            [
                Arg("name", help="The name of the deployment"),
                Group(
                    Arg("--csv", action="store_true", help="print as CSV"),
                    Arg("--json", action="store_true", help="print as JSON"),
                    Arg("--yaml", action="store_true", help="print as YAML"),
                ),
            ],
        ),
    ],
)

args_description = [main_cmd]  # type: List[Any]



# © Copyright 2024 Hewlett Packard Enterprise Development LP
from argparse import Namespace
from typing import Any, List

import aiolirest
from aioli import cli
from aioli.cli import render
from aioli.common.api import authentication
from aioli.common.declarative_argparse import Arg, Cmd, Group
from aiolirest.models.role_assignment import RoleAssignment
from aiolirest.models.role_assignments import RoleAssignments


@authentication.required
def list_roles(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.RolesApi(session)
        response = api_instance.roles_get()

    headers = [
        "ID",
        "Role Name",
    ]

    if args.json:
        render.print_json([r.to_dict() for r in response])
    elif args.yaml:
        print(render.format_object_as_yaml([r.to_dict() for r in response]))
    else:
        values = [[r.id, r.role_name] for r in response]
        render.tabulate_or_csv(headers, values, args.csv)


@authentication.required
def list_user_roles(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.RolesApi(session)
        response = api_instance.roles_assignments_get(args.username)

    headers = [
        "User Name",
        "Role Name",
    ]

    if args.json:
        render.print_json([r.to_dict() for r in response])
    elif args.yaml:
        print(render.format_object_as_yaml([r.to_dict() for r in response]))
    else:
        values = [[r.user_name, r.role_name] for r in response]
        render.tabulate_or_csv(headers, values, args.csv)


@authentication.required
def assign_role(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.RolesApi(session)

    assignments = []
    assignments.append(RoleAssignment(userName=args.u, roleName=args.role))

    api_instance.roles_add_assignments_post(RoleAssignments(userRoleAssignments=assignments))


@authentication.required
def unassign_role(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.RolesApi(session)

    assignments = []
    assignments.append(RoleAssignment(userName=args.u, roleName=args.role))

    api_instance.roles_remove_assignments_post(RoleAssignments(userRoleAssignments=assignments))


main_cmd = Cmd(
    "rbac",
    None,
    "role management",
    [
        Cmd(
            "list-roles ls",
            list_roles,
            "list roles",
            [
                Group(
                    Arg("--csv", action="store_true", help="print as CSV"),
                    Arg("--json", action="store_true", help="print as JSON"),
                    Arg("--yaml", action="store_true", help="print as YAML"),
                ),
            ],
            is_default=True,
        ),
        Cmd(
            "list-user-roles lu",
            list_user_roles,
            "list roles for a user",
            [
                Arg("username", help="The name of the user"),
                Group(
                    Arg("--csv", action="store_true", help="print as CSV"),
                    Arg("--json", action="store_true", help="print as JSON"),
                    Arg("--yaml", action="store_true", help="print as YAML"),
                ),
            ],
        ),
        Cmd(
            "assign-role",
            assign_role,
            "assign a role to a user",
            [
                Arg(
                    "role",
                    help="role name",
                ),
                Arg("-u", help="The name of the user", required="true"),
            ],
        ),
        Cmd(
            "unassign-role",
            unassign_role,
            "unassign a role for a user",
            [
                Arg(
                    "role",
                    help="role name",
                ),
                Arg("-u", help="The name of the user", required="true"),
            ],
        ),
    ],
)


args_description = [main_cmd]  # type: List[Any]



# coding: utf-8

"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictStr
from pydantic import Field
from aiolirest.models.resource_profile import ResourceProfile
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class ConfigurationResources(BaseModel):
    """
    ConfigurationResources models a configuration's resource requirements.
    """ # noqa: E501
    gpu_type: Optional[StrictStr] = Field(default=None, description="An optional GPU type required by this service (e.g. nvidia-tesla-a100). If unspecified, any GPU available will be used. Your cluster administrator should provide the list of GPU types that can be specified on your cluster. Unless otherwise configured by the administrator, the default resource name for GPU allocation is nvidia.com/gpu. You can override this default resource name for GPUs by specifying an alternate resource name here (it will include a / character). Your administrator may provide other resource names for GPUs available on your cluster (e.g. amd.com/gpu, intel.com/gpu, or a MIG format such as nvidia.com/mig-1g.5gb).", alias="gpuType")
    limits: Optional[ResourceProfile] = None
    requests: Optional[ResourceProfile] = None
    __properties: ClassVar[List[str]] = ["gpuType", "limits", "requests"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of ConfigurationResources from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of limits
        if self.limits:
            _dict['limits'] = self.limits.to_dict()
        # override the default output from pydantic by calling `to_dict()` of requests
        if self.requests:
            _dict['requests'] = self.requests.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of ConfigurationResources from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "gpuType": obj.get("gpuType"),
            "limits": ResourceProfile.from_dict(obj.get("limits")) if obj.get("limits") is not None else None,
            "requests": ResourceProfile.from_dict(obj.get("requests")) if obj.get("requests") is not None else None
        })
        return _obj





# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
class FeatureFlagDisabled(Exception):
    """
    Exception indicating that there is a currently disabled feature flag
    that is required to use a feature
    """

    pass


class CliError(Exception):
    """
    Base class for all CLI errors.
    """

    name: str

    def __init__(self, message: str, exit_code: int = 1) -> None:
        """
        Args:
        - e_stack: The exception that triggered this error.
        - exit_code: The exit code to use when exiting the CLI.
        """
        super().__init__(message)
        self.name = "Error"
        self.exit_code = exit_code
        self.message = message



# coding: utf-8

"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictInt, StrictStr
from pydantic import Field
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class Autoscaling(BaseModel):
    """
    Configuration that controls scaling up/down.
    """ # noqa: E501
    max_replicas: Optional[StrictInt] = Field(default=None, description="Maximum number of replicas created based upon demand.", alias="maxReplicas")
    metric: Optional[StrictStr] = Field(default='concurrency', description="Metric name which controls autoscaling. Must be one of the values: (concurrency, rps, cpu, memory).")
    min_replicas: Optional[StrictInt] = Field(default=None, description="Minimum number of replicas.   A value of zero, enables scale-down-to-zero for the service. By default, scale-down happens when no requests have occurred for more than 30 seconds. Requests that occur when there are zero replicas may experience longer delays while the container is started.   During this time the request is blocked. On rollout of a new service version, a value greater than zero will cause the prior version to remain in standby until garbage collected (up to 24 hours).", alias="minReplicas")
    target: Optional[StrictInt] = Field(default=None, description="Metric target value.  When the metric value reaches the target, additional replicas will be created. A concurrency target of 1, will cause a second replica to be started whenever there is an active request.")
    __properties: ClassVar[List[str]] = ["maxReplicas", "metric", "minReplicas", "target"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of Autoscaling from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of Autoscaling from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "maxReplicas": obj.get("maxReplicas"),
            "metric": obj.get("metric") if obj.get("metric") is not None else 'concurrency',
            "minReplicas": obj.get("minReplicas"),
            "target": obj.get("target")
        })
        return _obj





# © Copyright 2024 Hewlett Packard Enterprise Development LP
import json as _json
import os
import types
import webbrowser
from typing import Any, Dict, Iterator, Optional, Tuple, Union
from urllib import parse

import requests
import urllib3

import aioli
import aioli.common.requests
from aioli.common.api import authentication, certs, errors


def parse_master_address(master_address: str) -> parse.ParseResult:
    if master_address.startswith("https://"):
        default_port = 443
    elif master_address.startswith("http://"):
        default_port = 80
    else:
        default_port = 8080
        master_address = "http://{}".format(master_address)
    parsed = parse.urlparse(master_address)
    if not parsed.port:
        parsed = parsed._replace(netloc="{}:{}".format(parsed.netloc, default_port))
    return parsed


def make_url(master_address: str, suffix: str) -> str:
    """@deprecated use make_url_new instead"""
    parsed = parse_master_address(master_address)
    return parse.urljoin(parsed.geturl(), suffix)


def make_url_new(master_address: str, suffix: str) -> str:
    parsed_suffix = parse.urlparse(suffix)
    if parsed_suffix.scheme and parsed_suffix.netloc:
        return make_url(master_address, suffix)
    parsed = parse_master_address(master_address)
    master_url = parsed.geturl().rstrip("/")
    suffix = suffix.lstrip("/")
    separator = "/" if suffix or master_address.endswith("/") else ""
    return "{}{}{}".format(master_url, separator, suffix)


def maybe_upgrade_ws_scheme(master_address: str) -> str:
    parsed = parse.urlparse(master_address)
    if parsed.scheme == "https":
        return parsed._replace(scheme="wss").geturl()
    elif parsed.scheme == "http":
        return parsed._replace(scheme="ws").geturl()
    else:
        return master_address


def make_interactive_task_url(
    task_id: str,
    service_address: str,
    description: str,
    resource_pool: str,
    task_type: str,
    currentSlotsExceeded: bool,
) -> str:
    wait_path = (
        "/jupyter-lab/{}/events".format(task_id)
        if task_type == "jupyter-lab"
        else "/tensorboard/{}/events?tail=1".format(task_id)
    )
    wait_path_url = service_address + wait_path
    public_url = os.environ.get("PUBLIC_URL", "/")
    wait_page_url = "{}/wait/{}/{}?eventUrl={}&serviceAddr={}".format(
        public_url, task_type, task_id, wait_path_url, service_address
    )
    task_web_url = "{}/interactive/{}/{}/{}/{}/{}?{}".format(
        public_url,
        task_id,
        task_type,
        parse.quote(description),
        resource_pool,
        parse.quote_plus(wait_page_url),
        f"currentSlotsExceeded={str(currentSlotsExceeded).lower()}",
    )
    return task_web_url


def do_request(
    method: str,
    host: str,
    path: str,
    params: Optional[Dict[str, Any]] = None,
    json: Any = None,
    data: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    authenticated: bool = True,
    auth: Optional[authentication.Authentication] = None,
    cert: Optional[certs.Cert] = None,
    stream: bool = False,
    timeout: Optional[Union[Tuple, float]] = None,
    max_retries: Optional[urllib3.util.retry.Retry] = None,
) -> requests.Response:
    if headers is None:
        h: Dict[str, str] = {}
    else:
        h = headers

    if cert is None:
        cert = certs.cli_cert

    # set the token and username based on this order:
    # - argument `auth`
    # - header `Authorization`
    # - existing cli_auth
    # - allocation_token

    username = ""
    if auth is not None:
        if authenticated:
            h["Authorization"] = "Bearer {}".format(auth.get_session_token())
        username = auth.get_session_user()
    elif h.get("Authorization") is not None:
        pass
    elif authentication.cli_auth is not None:
        if authenticated:
            h["Authorization"] = "Bearer {}".format(authentication.cli_auth.get_session_token())
        username = authentication.cli_auth.get_session_user()
    # elif authenticated and h.get("Grpc-Metadata-x-allocation-token") is None:
    #     allocation_token = authentication.get_allocation_token()
    #     if allocation_token:
    #         h["Grpc-Metadata-x-allocation-token"] = "Bearer {}".format(allocation_token)

    if params is None:
        params = {}

    # Allow the json json to come pre-encoded, if we need custom encoding.
    if json is not None and data is not None:
        raise ValueError("json and data must not be provided together")

    if json:
        data = aioli.util.json_encode(json)

    try:
        r = aioli.common.requests.request(
            method,
            make_url(host, path),
            params=params,
            data=data,
            headers=h,
            verify=cert.bundle if cert else None,
            stream=stream,
            timeout=timeout,
            server_hostname=cert.name if cert else None,
            max_retries=max_retries,
        )
    except requests.exceptions.SSLError:
        raise
    except requests.exceptions.ConnectionError as e:
        raise errors.MasterNotFoundException(str(e))
    except requests.exceptions.RequestException as e:
        raise errors.BadRequestException(str(e))

    def _get_error_str(r: requests.models.Response) -> str:
        try:
            json_resp = _json.loads(r.text)
            mes = json_resp.get("message")
            if mes is not None:
                return str(mes)
            # Try getting GRPC error description if message does not exist.
            return str(json_resp.get("error").get("error"))
        except Exception:
            return ""

    if r.status_code == 403:
        raise errors.ForbiddenException(username=username, message=_get_error_str(r))
    if r.status_code == 401:
        raise errors.UnauthenticatedException(username=username)
    elif r.status_code == 404:
        raise errors.NotFoundException(_get_error_str(r))
    elif r.status_code >= 300:
        raise errors.APIException(r)

    return r


def get(
    host: str,
    path: str,
    params: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, str]] = None,
    authenticated: bool = True,
    auth: Optional[authentication.Authentication] = None,
    cert: Optional[certs.Cert] = None,
    stream: bool = False,
    timeout: Optional[Union[Tuple, float]] = None,
) -> requests.Response:
    """
    Send a GET request to the remote API.
    """
    return do_request(
        "GET",
        host,
        path,
        params=params,
        headers=headers,
        authenticated=authenticated,
        auth=auth,
        cert=cert,
        stream=stream,
    )


def delete(
    host: str,
    path: str,
    params: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, str]] = None,
    authenticated: bool = True,
    auth: Optional[authentication.Authentication] = None,
    cert: Optional[certs.Cert] = None,
    timeout: Optional[Union[Tuple, float]] = None,
) -> requests.Response:
    """
    Send a DELETE request to the remote API.
    """
    return do_request(
        "DELETE",
        host,
        path,
        params=params,
        headers=headers,
        authenticated=authenticated,
        auth=auth,
        cert=cert,
        timeout=timeout,
    )


def post(
    host: str,
    path: str,
    json: Any = None,
    headers: Optional[Dict[str, str]] = None,
    authenticated: bool = True,
    auth: Optional[authentication.Authentication] = None,
    cert: Optional[certs.Cert] = None,
    timeout: Optional[Union[Tuple, float]] = None,
) -> requests.Response:
    """
    Send a POST request to the remote API.
    """
    return do_request(
        "POST",
        host,
        path,
        json=json,
        headers=headers,
        authenticated=authenticated,
        auth=auth,
        cert=cert,
        timeout=timeout,
    )


def patch(
    host: str,
    path: str,
    json: Dict[str, Any],
    headers: Optional[Dict[str, str]] = None,
    authenticated: bool = True,
    auth: Optional[authentication.Authentication] = None,
    cert: Optional[certs.Cert] = None,
    timeout: Optional[Union[Tuple, float]] = None,
) -> requests.Response:
    """
    Send a PATCH request to the remote API.
    """
    return do_request(
        "PATCH",
        host,
        path,
        json=json,
        headers=headers,
        authenticated=authenticated,
        auth=auth,
        cert=cert,
        timeout=timeout,
    )


def put(
    host: str,
    path: str,
    json: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, str]] = None,
    authenticated: bool = True,
    auth: Optional[authentication.Authentication] = None,
    cert: Optional[certs.Cert] = None,
    timeout: Optional[Union[Tuple, float]] = None,
) -> requests.Response:
    """
    Send a PUT request to the remote API.
    """
    return do_request(
        "PUT",
        host,
        path,
        json=json,
        headers=headers,
        authenticated=authenticated,
        auth=auth,
        cert=cert,
        timeout=timeout,
    )


def browser_open(host: str, path: str) -> str:
    url = make_url(host, path)
    webbrowser.open(url)
    return url


class WebSocket:
    def __init__(self, socket: Any) -> None:
        import lomond

        self.socket = socket  # type: lomond.WebSocket

    def __enter__(self) -> "WebSocket":
        return self

    def __iter__(self) -> Iterator[Any]:
        from lomond import events

        for event in self.socket.connect(ping_rate=0):
            if isinstance(event, events.Connected):
                # Ignore the initial connection event.
                pass
            elif isinstance(event, (events.Closing, events.Disconnected)):
                # The socket was successfully closed so we just return.
                return
            elif isinstance(
                event,
                (events.ConnectFail, events.Rejected, events.ProtocolError),
            ):
                # Any unexpected failures raise the standard API exception.
                raise errors.BadRequestException(message="WebSocket failure: {}".format(event))
            elif isinstance(event, events.Text):
                # All web socket connections are expected to be in a JSON
                # format.
                yield _json.loads(event.text)

    def __exit__(
        self,
        exc_type: Optional[type],
        exc_val: Optional[BaseException],
        exc_tb: Optional[types.TracebackType],
    ) -> None:
        if not self.socket.is_closed:
            self.socket.close()


def ws(host: str, path: str) -> WebSocket:
    """
    Connect to a web socket at the remote API.
    """
    import lomond

    websocket = lomond.WebSocket(maybe_upgrade_ws_scheme(make_url(host, path)))
    token = authentication.must_cli_auth().get_session_token()
    websocket.add_header("Authorization".encode(), "Bearer {}".format(token).encode())
    return WebSocket(websocket)



# coding: utf-8

# flake8: noqa
"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


# import models into model package
from aiolirest.models.auto_scaling_template import AutoScalingTemplate
from aiolirest.models.auto_scaling_template_request import AutoScalingTemplateRequest
from aiolirest.models.autoscaling import Autoscaling
from aiolirest.models.configuration_resources import ConfigurationResources
from aiolirest.models.controller_config import ControllerConfig
from aiolirest.models.controller_status import ControllerStatus
from aiolirest.models.deployment import Deployment
from aiolirest.models.deployment_model_version import DeploymentModelVersion
from aiolirest.models.deployment_priority import DeploymentPriority
from aiolirest.models.deployment_request import DeploymentRequest
from aiolirest.models.deployment_state import DeploymentState
from aiolirest.models.deployment_token import DeploymentToken
from aiolirest.models.deployment_token_patch_request import DeploymentTokenPatchRequest
from aiolirest.models.deployment_token_request import DeploymentTokenRequest
from aiolirest.models.ezua_info import EZUAInfo
from aiolirest.models.error_response import ErrorResponse
from aiolirest.models.event_info import EventInfo
from aiolirest.models.failure_info import FailureInfo
from aiolirest.models.login_request import LoginRequest
from aiolirest.models.login_response import LoginResponse
from aiolirest.models.model_auth_token import ModelAuthToken
from aiolirest.models.model_response import ModelResponse
from aiolirest.models.observability import Observability
from aiolirest.models.packaged_model import PackagedModel
from aiolirest.models.packaged_model_request import PackagedModelRequest
from aiolirest.models.project import Project
from aiolirest.models.project_request import ProjectRequest
from aiolirest.models.resource_profile import ResourceProfile
from aiolirest.models.resources_template import ResourcesTemplate
from aiolirest.models.resources_template_request import ResourcesTemplateRequest
from aiolirest.models.role import Role
from aiolirest.models.role_assignment import RoleAssignment
from aiolirest.models.role_assignments import RoleAssignments
from aiolirest.models.security import Security
from aiolirest.models.success_response import SuccessResponse
from aiolirest.models.telemetry_info import TelemetryInfo
from aiolirest.models.trained_model_registry import TrainedModelRegistry
from aiolirest.models.trained_model_registry_request import TrainedModelRegistryRequest
from aiolirest.models.user import User
from aiolirest.models.user_patch_request import UserPatchRequest
from aiolirest.models.user_request import UserRequest
from aiolirest.models.user_session import UserSession



# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
import hashlib
import json
import os
import socket
import ssl
import sys
import urllib.parse
from argparse import ArgumentDefaultsHelpFormatter, ArgumentError, ArgumentParser
from typing import List, Optional, Sequence, Union, cast

import argcomplete
import argcomplete.completers
from OpenSSL import SSL, crypto
from termcolor import colored
from urllib3.exceptions import MaxRetryError, SSLError

import aioli
from aioli.cli import render
from aioli.cli.deployment import args_description as deployment_args_description
from aioli.cli.model import args_description as model_args_description
from aioli.cli.project import args_description as project_args_description
from aioli.cli.registry import args_description as registry_args_description
from aioli.cli.role import args_description as role_args_description
from aioli.cli.sso import args_description as sso_args_description
from aioli.cli.template import args_description as template_args_description
from aioli.cli.token import args_description as deployment_token_args_description
from aioli.cli.user import args_description as user_args_description
from aioli.cli.version import args_description as version_args_description
from aioli.cli.version import check_version
from aioli.common import api
from aioli.common.api import certs
from aioli.common.check import check_not_none
from aioli.common.declarative_argparse import Arg, ArgsDescription, add_args
from aioli.common.util import chunks, debug_mode, get_default_controller_address
from aiolirest.rest import ApiException

from .errors import CliError, FeatureFlagDisabled

args_description: ArgsDescription = [
    Arg("-u", "--user", help="run as the given user", metavar="username", default=None),
    Arg(
        "-c",
        "--controller",
        help="controller address",
        metavar="address",
        default=get_default_controller_address(),
    ),
    Arg(
        "-v",
        "--version",
        action="version",
        help="print CLI version and exit",
        version="%(prog)s {}".format(aioli.__version__),
    ),
]
all_args_description: ArgsDescription = (
    args_description
    + project_args_description
    + registry_args_description
    + model_args_description
    + deployment_args_description
    + version_args_description
    + user_args_description
    + role_args_description
    + sso_args_description
    + deployment_token_args_description
    + template_args_description
)


def make_parser() -> ArgumentParser:
    return ArgumentParser(
        description="Aioli command-line client",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )


def die(message: str, always_print_traceback: bool = False, exit_code: int = 1) -> None:
    if always_print_traceback or debug_mode():
        import traceback

        traceback.print_exc(file=sys.stderr)

    print(colored(message, "red"), file=sys.stderr, end="\n")
    exit(exit_code)


def main(
    args: List[str] = sys.argv[1:],
) -> None:
    if sys.platform == "win32":
        # Magic incantation to make a Windows 10 cmd.exe process color-related ANSI escape codes.
        os.system("")

    # we lazily import "det deploy" but in the future we'd want to lazily import everything.
    parser = make_parser()
    add_args(parser, all_args_description)

    try:
        argcomplete.autocomplete(parser)

        parsed_args = parser.parse_args(args)
        parsed_args.controller = parsed_args.controller.rstrip("/")
        url = urllib.parse.urlparse(parsed_args.controller)
        if url.netloc:
            parsed_args.controller = url.scheme + "://" + url.netloc

        v = vars(parsed_args)
        if not v.get("func"):
            parser.print_usage()
            parser.exit(2, "{}: no subcommand specified\n".format(parser.prog))

        try:
            configure_certificate_for_controller(parsed_args.controller)
            parsed_args.func(parsed_args)
        except KeyboardInterrupt as e:
            raise e
        except (
            api.errors.BadRequestException,
            api.errors.BadResponseException,
            MaxRetryError,
        ) as e:
            die(f"Failed to {parsed_args.func.__name__}: {e}")
        except api.errors.CorruptTokenCacheException:
            die(
                "Failed to login: Attempted to read a corrupted token cache. "
                "The store has been deleted; please try again."
            )
        except FeatureFlagDisabled as e:
            die(f"controller does not support this operation: {e}")
        except CliError as e:
            die(e.message, exit_code=e.exit_code)
        except ArgumentError as e:
            die(e.message, exit_code=2)
        except ApiException as e:
            if "Client sent an HTTP request to an HTTPS server" in e.body:
                die(f"Ensure the controller address includes 'https://': {e.reason}: {e.body}")
            else:
                message = extract_message(e)
                if message is not None and len(message) > 1:
                    # The first character of the message may/should be capitalized.
                    # In this context however it looks better if in lower case.
                    message = message[0].lower() + message[1 : len(message)]
                    die(f"Failed to {parsed_args.func.__name__}: {message}")

                die(
                    f"Failed on REST API operation, status {e.status}, reason: "
                    f"{e.reason}, body: {e.body}"
                )
        except Exception:
            die(f"Failed to {parsed_args.func.__name__}", always_print_traceback=True)
    except KeyboardInterrupt:
        die("Interrupting...", exit_code=3)


def configure_certificate_for_controller(controller: str) -> None:
    # Configure the CLI's Cert singleton.
    certs.cli_cert = certs.default_load(controller)

    try:
        # check_version doesn't require credentials, so we can use it to verify the certificate.
        check_version(controller)
    except SSLError:
        # An SSLError usually means that we queried a controller over HTTPS and got an
        # untrusted cert, so allow the user to store and trust the current cert. (It
        # could also mean that we tried to talk HTTPS on the HTTP port, but distinguishing
        # that based on the exception is annoying, and we'll figure that out in the next
        # step anyway.)
        addr = api.parse_master_address(controller)
        check_not_none(addr.hostname)
        check_not_none(addr.port)
        try:
            ctx = SSL.Context(SSL.TLSv1_2_METHOD)
            conn = SSL.Connection(ctx, socket.socket())
            conn.set_tlsext_host_name(cast(str, addr.hostname).encode())
            conn.connect(cast(Sequence[Union[str, int]], (addr.hostname, addr.port)))
            conn.do_handshake()
            peer_cert_chain = conn.get_peer_cert_chain()
            if peer_cert_chain is None or len(peer_cert_chain) == 0:
                # Peer presented no cert.  It seems unlikely that this is possible after
                # do_handshake() succeeded, but checking for None makes mypy happy.
                raise crypto.Error()
            cert_pem_data = [
                crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode()
                for cert in peer_cert_chain
            ]
        except crypto.Error:
            die(
                "Tried to connect over HTTPS but couldn't get a certificate from the "
                "controller; consider using HTTP"
            )

        # Compute the fingerprint of the certificate; this is the same as the output of
        # `openssl x509 -fingerprint -sha256 -inform pem -noout -in <cert>`.
        cert_hash = hashlib.sha256(ssl.PEM_cert_to_DER_cert(cert_pem_data[0])).hexdigest()
        cert_fingerprint = ":".join(chunks(cert_hash, 2))

        if not render.yes_or_no(
            "The controller sent an untrusted certificate chain with this SHA256 "
            "fingerprint:\n"
            "{}\nDo you want to trust this certificate from now on?".format(cert_fingerprint)
        ):
            die("Unable to verify controller certificate")

        joined_certs = "".join(cert_pem_data)
        certs.CertStore(certs.default_store()).set_cert(controller, joined_certs)
        # Reconfigure the CLI's Cert singleton, but preserve the certificate name.
        old_cert_name = certs.cli_cert.name
        certs.cli_cert = certs.Cert(cert_pem=joined_certs, name=old_cert_name)

        try:
            # let's try that again now that we have reconfigured with the certificate
            # information supplied by the controller.
            check_version(controller)
        except SSLError:
            die("Failed to verify the controller certificate")


def extract_message(e: ApiException) -> Optional[str]:
    decoder = json.JSONDecoder()
    try:
        body_obj = decoder.decode(e.body)
        return str(body_obj["message"])
    except json.JSONDecodeError:
        return None



# coding: utf-8

"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictInt, StrictStr
from pydantic import Field
from aiolirest.models.failure_info import FailureInfo
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class DeploymentState(BaseModel):
    """
    Describes the current state of the deployment as managed by KServe
    """ # noqa: E501
    endpoint: Optional[StrictStr] = Field(default=None, description="Endpoint to access inference service.")
    failure_info: Optional[List[FailureInfo]] = Field(default=None, description="List of any failures that have occurred.", alias="failureInfo")
    mdl_id: Optional[StrictStr] = Field(default=None, description="The ID of the deployed packaged model associated with this state.", alias="modelId")
    native_app_name: Optional[StrictStr] = Field(default=None, description="The name of the Kubernetes application for the specific service version. Use this name to match the app value in Grafana/Prometheus to obtain logs and metrics for this deployed service version.", alias="nativeAppName")
    status: Optional[StrictStr] = Field(default=None, description="Status of a particular service revision. * `Deploying` - Service configuration is in progress. * `Failed` - The service configuration failed. * `Ready` - The service has been successfully configured and is serving. * `Updating` - A new service revision is being rolled-out. * `UpdateFailed` - The current service revision failed to rollout due to an error.   The prior version is still serving requests. * `Deleting` - The deployed service is being removed. * `Paused` - The deployed service has been stopped by the user or an external action. * `Unknown` - Unable to determined the status. * `Canceled` - The specified model version of the deployment was canceled by the user.  The inference service is still active, but serving no requests.  This is a read-only property. Must be one of the values: (Deploying,Ready,Updating,UpdateFailed,Failed,Deleting,Paused,Unknown).")
    traffic_percentage: Optional[StrictInt] = Field(default=None, description="Target percentage of traffic intended for this service version when successfully deployed.", alias="trafficPercentage")
    __properties: ClassVar[List[str]] = ["endpoint", "failureInfo", "modelId", "nativeAppName", "status", "trafficPercentage"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of DeploymentState from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in failure_info (list)
        _items = []
        if self.failure_info:
            for _item in self.failure_info:
                if _item:
                    _items.append(_item.to_dict())
            _dict['failureInfo'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of DeploymentState from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "endpoint": obj.get("endpoint"),
            "failureInfo": [FailureInfo.from_dict(_item) for _item in obj.get("failureInfo")] if obj.get("failureInfo") is not None else None,
            "modelId": obj.get("modelId"),
            "nativeAppName": obj.get("nativeAppName"),
            "status": obj.get("status"),
            "trafficPercentage": obj.get("trafficPercentage")
        })
        return _obj





# flake8: noqa

# import apis into api package
from aiolirest.api.authentication_api import AuthenticationApi
from aiolirest.api.deployments_api import DeploymentsApi
from aiolirest.api.information_api import InformationApi
from aiolirest.api.packaged_models_api import PackagedModelsApi
from aiolirest.api.projects_api import ProjectsApi
from aiolirest.api.registries_api import RegistriesApi
from aiolirest.api.roles_api import RolesApi
from aiolirest.api.templates_api import TemplatesApi
from aiolirest.api.tokens_api import TokensApi
from aiolirest.api.users_api import UsersApi




# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
import enum


class PageOpts(str, enum.Enum):
    single = "1"
    all = "all"


# HTTP status codes that will force request retries.
# Bad Gateway, Service Unavailable, Gateway Timeout
RETRY_STATUSES = [502, 503, 504]

# Default max number of times to retry a request.
MAX_RETRIES = 5



# © Copyright 2024 Hewlett Packard Enterprise Development LP
import argparse
import contextlib
import functools
import getpass
import hashlib
import json
import pathlib
import urllib.parse
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    cast,
)

import filelock
import requests

import aioli.common.api.errors
import aiolirest
from aioli import cli
from aioli.common import api, util
from aioli.common.api import certs

Credentials = NamedTuple("Credentials", [("username", str), ("password", str)])

PASSWORD_SALT = "GubPEmmotfiK9TMD6Zdw"


def salt_and_hash(password: str) -> str:
    if password:
        return hashlib.sha512((PASSWORD_SALT + password).encode()).hexdigest()
    else:
        return password


class UsernameTokenPair:
    def __init__(self, username: str, token: str):
        self.username = username
        self.token = token


def default_load_user_password(
    requested_user: Optional[str],
    password: Optional[str],
    token_store: "TokenStore",
) -> Tuple[Optional[str], Optional[str]]:
    # Always prefer an explicitly provided user/password.
    if requested_user:
        return requested_user, password

    # Next highest priority is user/password from environment.
    # Watch out! We have to check for AIOLI_USER and AIOLI_PASS, because containers will
    # have AIOLI_USER set, but that doesn't overrule the active user in the TokenStore, because
    # if the TokenStore in the container has an active user, that means the user has explicitly
    # ran `aioli user login` inside the container.
    if (
        util.get_aioli_username_from_env() is not None
        and util.get_aioli_password_from_env() is not None
    ):
        return util.get_aioli_username_from_env(), util.get_aioli_password_from_env()

    # Last priority is the active user in the token store.
    return token_store.get_active_user(), password


def get_rest_config(host: str) -> aiolirest.Configuration:
    cert = certs.get_cert_path()
    if cert is not None and cert is not False:
        config = aiolirest.Configuration(host=host, ssl_ca_cert=cert)
    else:
        config = aiolirest.Configuration(host=host)

    if cert is False:
        config.verify_ssl = False

    url_parsed = urllib.parse.urlparse(host)
    proxies = requests.utils.getproxies()  # type: ignore[attr-defined]
    proxy_url = proxies.get("https") if host.startswith("https") else proxies.get("http", None)
    if proxy_url is not None:
        if not requests.utils.proxy_bypass(url_parsed.netloc):  # type: ignore[attr-defined]
            config.proxy = proxy_url

    return config


class Authentication:
    def __init__(
        self,
        master_address: Optional[str] = None,
        requested_user: Optional[str] = None,
        password: Optional[str] = None,
        cert: Optional[certs.Cert] = None,
    ) -> None:
        self.master_address = master_address or util.get_default_controller_address()
        self.token_store = TokenStore(self.master_address)

        self.session = self._init_session(requested_user, password, cert)

    def _init_session(
        self,
        requested_user: Optional[str],
        password: Optional[str],
        cert: Optional[certs.Cert],
    ) -> UsernameTokenPair:
        # Get session_user and password given the following priority:
        # 1. User passed in with flag (requested_user)
        # 2. User from environment if AIOLI_PASS is set.
        # 3. Active user from the token store.
        session_user, password = default_load_user_password(
            requested_user, password, self.token_store
        )

        if session_user is None:
            raise aioli.common.api.errors.UnauthenticatedException(username="")

        # Check the token store if this session_user has a cached token. If so, check with the
        # master to verify it has not expired. Otherwise, let the token be None.
        token = self.token_store.get_token(session_user)
        if token is not None and not _is_token_valid(self.master_address, token, cert):
            self.token_store.drop_user(session_user)
            token = None

        # Special case: use token provided from the container environment if:
        # - No token was obtained from the token store already,
        # - There is a token available from the container environment, and
        # - No user was explicitly requested, or the requested user matches the token available
        #   in the container environment.
        if (
            token is None
            and util.get_aioli_username_from_env() is not None
            and util.get_aioli_user_token_from_env() is not None
            and requested_user in (None, util.get_aioli_username_from_env())
        ):
            session_user = util.get_aioli_username_from_env()
            assert session_user
            token = util.get_aioli_user_token_from_env()

        if token is not None:
            return UsernameTokenPair(session_user, token)

        if password is None:
            password = getpass.getpass("Password for user '{}': ".format(session_user))

        try:
            token = do_login(self.master_address, session_user, salt_and_hash(password), True, cert)
        except api.errors.ForbiddenException:
            raise

        self.token_store.set_token(session_user, token)

        return UsernameTokenPair(session_user, token)

    def is_user_active(self, username: str) -> bool:
        return self.token_store.get_active_user() == username

    def get_session_user(self) -> str:
        """
        Returns the session user for the current session. If there is no active
        session, then an UnauthenticatedException will be raised.
        """
        return self.session.username

    def get_session_token(self, must: bool = True) -> str:
        """
        Returns the authentication token for the session user. If there is no
        active session, then an UnauthenticatedException will be raised.
        """
        if self.session is None:
            if must:
                raise api.errors.UnauthenticatedException(username="")
            else:
                return ""
        return self.session.token


def do_login(
    controller_address: str,
    username: str,
    password: str,
    isHashed: bool,
    cert: Optional[certs.Cert] = None,
) -> str:
    host: str = controller_address
    host = util.prepend_protocol(host)

    client = aiolirest.ApiClient(get_rest_config(host))
    request = aiolirest.LoginRequest(username=username, password=password, isHashed=isHashed)
    response = aiolirest.AuthenticationApi(client).login_post(request=request)

    token = response.token

    return str(token)


class LogoutAuthentication(Authentication):
    """
    An api-compatible Authentication object that is basically exactly a UserTokenPair.

    TODO(MLG-215): delete Authentication class and write a function that returns a UsernameTokenPair
    in its place, and let do_request() take UsernameTokenPair as input.
    """

    def __init__(self, session_user: str, session_token: str) -> None:
        self.session_user = session_user
        self.session_token = session_token

    def get_session_user(self) -> str:
        return self.session_user

    def get_session_token(self, must: bool = True) -> str:
        return self.session_token


def logout(
    controller_address: Optional[str],
    requested_user: Optional[str],
    cert: Optional[certs.Cert],
) -> None:
    """
    Logout if there is an active session for this master/username pair, otherwise do nothing.

    Additionally, if the user happens to be the active user, drop the active user from the
    TokenStore.
    """

    controller_address = controller_address or util.get_default_controller_address()
    token_store = TokenStore(controller_address)

    session_user, _ = default_load_user_password(requested_user, None, token_store)

    if session_user is None:
        return

    if session_user == token_store.get_active_user():
        token_store.clear_active()

    session_token = token_store.get_token(session_user)

    if session_token is None:
        return

    token_store.drop_user(session_user)

    null_namespace = argparse.Namespace()

    with cli.setup_session(args=null_namespace, controller=controller_address) as session:
        api_instance = aiolirest.AuthenticationApi(session)
        api_instance.logout_post()


def logout_all(master_address: Optional[str], cert: Optional[certs.Cert]) -> None:
    master_address = master_address or util.get_default_controller_address()
    token_store = TokenStore(master_address)

    users = token_store.get_all_users()

    for user in users:
        logout(master_address, user, cert)

    token_store.clear_active()


def _is_token_valid(controller_address: str, token: str, cert: Optional[certs.Cert]) -> bool:
    """
    Find out whether the given token is valid by attempting to use it
    on the "api/v1/me" endpoint.
    """
    host: str = controller_address
    host = util.prepend_protocol(host)

    configuration = get_rest_config(host)
    configuration.api_key["ApiKeyAuth"] = "Bearer " + token
    client = aiolirest.ApiClient(configuration)
    users_api = aiolirest.UsersApi(client)

    try:
        users_api.users_me_get()
    except (api.errors.UnauthenticatedException, api.errors.APIException):
        return False
    return True


class TokenStore:
    """
    TokenStore is a class for reading/updating a persistent store of user authentication tokens.
    TokenStore can remember tokens for many users for each of many masters.  It can also remembers
    one "active user" for each master, which is set via `aioli user login`.

    All updates to the file follow a read-modify-write pattern, and use file locks to protect the
    integrity of the underlying file cache.
    """

    def __init__(self, master_address: str, path: Optional[pathlib.Path] = None) -> None:
        self.master_address = master_address
        self.path = path or util.get_config_path().joinpath("auth.json")
        self.path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
        # Decide on paths for a lock file and a temp files (during writing)
        self.temp = pathlib.Path(str(self.path) + ".temp")
        self.lock = str(self.path) + ".lock"

        with filelock.FileLock(self.lock):
            store = self._load_store_file()

        self._reconfigure_from_store(store)

    def _reconfigure_from_store(self, store: dict) -> None:
        substore = store.get("masters", {}).get(self.master_address, {})
        self._active_user = cast(str, substore.get("active_user"))
        self._tokens = cast(Dict[str, str], substore.get("tokens", {}))

    def get_active_user(self) -> Optional[str]:
        return self._active_user

    def get_all_users(self) -> List[str]:
        return list(self._tokens)

    def get_token(self, user: str) -> Optional[str]:
        token = self._tokens.get(user)
        if token is not None:
            assert isinstance(token, str), "invalid cache; token must be a string"
        return token

    def delete_token_cache(self) -> None:
        with filelock.FileLock(self.lock):
            if self.path.exists():
                self.path.unlink()

    def drop_user(self, username: str) -> None:
        with self._persistent_store() as substore:
            tokens = substore.setdefault("tokens", {})
            if username in tokens:
                del tokens[username]

    def set_token(self, username: str, token: str) -> None:
        with self._persistent_store() as substore:
            tokens = substore.setdefault("tokens", {})
            tokens[username] = token

    def set_active(self, username: str) -> None:
        with self._persistent_store() as substore:
            tokens = substore.setdefault("tokens", {})
            if username not in tokens:
                raise api.errors.UnauthenticatedException(username=username)
            substore["active_user"] = username

    def clear_active(self) -> None:
        with self._persistent_store() as substore:
            substore.pop("active_user", None)

    @contextlib.contextmanager
    def _persistent_store(self) -> Iterator[Dict[str, Any]]:
        """
        Yields the appropriate store[self.master_address] that can be modified, and the modified
        result will be written back to file.

        Whatever updates are made will also be updated on self automatically.
        """
        with filelock.FileLock(self.lock):
            store = self._load_store_file()
            substore = store.setdefault("masters", {}).setdefault(self.master_address, {})

            # No need for try/finally, because we don't update the file after failures.
            yield substore

            # Reconfigure our cached variables.
            self._reconfigure_from_store(store)

            with self.temp.open("w") as f:
                json.dump(store, f, indent=4, sort_keys=True)
            self.temp.replace(self.path)

    def _load_store_file(self) -> Dict[str, Any]:
        """
        Read a token store from a file, shimming it to the most recent version if necessary.

        If a v0 store is found it will be reconfigured as a v1 store based on the master_address
        that is being currently requested.
        """
        try:
            if not self.path.exists():
                return {"version": 1}

            try:
                with self.path.open() as f:
                    store = json.load(f)
            except json.JSONDecodeError:
                raise api.errors.CorruptTokenCacheException()

            if not isinstance(store, dict):
                raise api.errors.CorruptTokenCacheException()

            version = store.get("version", 0)
            if version == 0:
                validate_token_store_v0(store)
                store = shim_store_v0(store, self.master_address)

            validate_token_store_v1(store)

            return cast(dict, store)

        except api.errors.CorruptTokenCacheException:
            # Delete invalid caches before exiting.
            self.path.unlink()
            raise


def shim_store_v0(v0: Dict[str, Any], master_address: str) -> Dict[str, Any]:
    """
    v1 schema is just a bit more nesting to support multiple masters.
    """
    v1 = {"version": 1, "masters": {master_address: v0}}
    return v1


def validate_token_store_v0(store: Any) -> bool:
    """
    Valid v0 schema example:

        {
          "active_user": "user_a",
          "tokens": {
            "user_a": "TOKEN",
            "user_b": "TOKEN"
          }
        }
    """

    if not isinstance(store, dict):
        raise api.errors.CorruptTokenCacheException()

    if len(set(store.keys()).difference({"active_user", "tokens"})) > 0:
        # Extra keys.
        raise api.errors.CorruptTokenCacheException()

    if "active_user" in store:
        if not isinstance(store["active_user"], str):
            raise api.errors.CorruptTokenCacheException()

    if "tokens" in store:
        tokens = store["tokens"]
        if not isinstance(tokens, dict):
            raise api.errors.CorruptTokenCacheException()
        for k, v in tokens.items():
            if not isinstance(k, str):
                raise api.errors.CorruptTokenCacheException()
            if not isinstance(v, str):
                raise api.errors.CorruptTokenCacheException()
    return True


def validate_token_store_v1(store: Any) -> bool:
    """
    Valid v1 schema example:

        {
          "version": 1,
          "masters": {
            "master_url_a": {
              "active_user": "user_a",
              "tokens": {
                "user_a": "TOKEN",
                "user_b": "TOKEN"
              }
            },
            "master_url_b": {
              "active_user": "user_c",
              "tokens": {
                "user_c": "TOKEN",
                "user_d": "TOKEN"
              }
            }
          }
        }

    Note that store["masters"] is a mapping of string url's to valid v0 schemas.
    """
    if not isinstance(store, dict):
        raise api.errors.CorruptTokenCacheException()

    if len(set(store.keys()).difference({"version", "masters"})) > 0:
        # Extra keys.
        raise api.errors.CorruptTokenCacheException()

    # Handle version.
    version = store.get("version")
    if version != 1:
        raise api.errors.CorruptTokenCacheException()

    if "masters" in store:
        masters = store["masters"]
        if not isinstance(masters, dict):
            raise api.errors.CorruptTokenCacheException()

        # Each entry of masters must be a master_url/substore pair.
        for key, val in masters.items():
            if not isinstance(key, str):
                raise api.errors.CorruptTokenCacheException()
            validate_token_store_v0(val)

    return True


# cli_auth is the process-wide authentication used for api calls originating from the cli.
cli_auth = None  # type: Optional[Authentication]


def required(func: Callable[[argparse.Namespace], Any]) -> Callable[..., Any]:
    """
    A decorator for cli functions.
    """

    @functools.wraps(func)
    def f(namespace: argparse.Namespace) -> Any:
        update_cli_auth(namespace)
        return func(namespace)

    return f


def must_cli_auth() -> Authentication:
    if not cli_auth:
        raise api.errors.UnauthenticatedException(username="")
    return cli_auth


def update_cli_auth(namespace: argparse.Namespace) -> None:
    global cli_auth
    cli_auth = Authentication(namespace.controller, namespace.user)



# © Copyright 2024 Hewlett Packard Enterprise Development LP
from argparse import Namespace
from typing import Any, Dict, List

import aiolirest
from aioli import cli
from aioli.cli import render
from aioli.common.api import authentication
from aioli.common.api.errors import NotFoundException
from aioli.common.declarative_argparse import Arg, Cmd, Group
from aiolirest.models.project import Project
from aiolirest.models.project_request import ProjectRequest


@authentication.required
def list_projects(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.ProjectsApi(session)
        response = api_instance.projects_get()

    if args.json:
        render.print_json(format_json(response))
    elif args.yaml:
        print(render.format_object_as_yaml(format_json(response)))
    else:
        headers = ["Name", "Description", "Owner"]
        values = [[p.name, p.description, p.owner] for p in response]
        values.sort()  # sort values by the first column
        render.tabulate_or_csv(headers, values, args.csv)


def format_json(response: List[Project]) -> List[Dict[str, str]]:
    return [project_to_dict(r) for r in response]


def project_to_dict(project: Project) -> Dict[str, str]:
    # Don't use the r.to_json() method as it adds backslash escapes for double quotes
    d: Dict[str, str] = project.to_dict()
    d.pop("id")
    d.pop("modifiedAt")
    return d


@authentication.required
def create_project(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.ProjectsApi(session)

        project_request = ProjectRequest(
            name=args.name,
            description=args.description,
            owner=args.owner,
        )
        api_instance.projects_post(project_request)


@authentication.required
def update_project(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.ProjectsApi(session)
        found = lookup_project(args.projectname, api_instance)
        request = ProjectRequest(
            name=found.name,
            description=found.description,
            owner=found.owner,
        )
        if args.name is not None:
            request.name = args.name

        if args.description is not None:
            request.description = args.description

        if args.owner is not None:
            request.owner = args.owner

        headers = {"Content-Type": "application/json"}
        assert found.id is not None
        api_instance.projects_id_put(found.id, request, _headers=headers)


@authentication.required
def delete_project(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.ProjectsApi(session)
        found = lookup_project(args.name, api_instance)
        assert found.id is not None
        api_instance.projects_id_delete(found.id)


@authentication.required
def show_project(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.ProjectsApi(session)

    project = lookup_project(args.name, api_instance)

    d = project.to_dict()
    if args.json:
        render.print_json(d)
    else:
        print(render.format_object_as_yaml(d))


def lookup_project(name: str, api: aiolirest.ProjectsApi) -> Project:
    for p in api.projects_get():
        if p.name == name:
            return p
    raise NotFoundException(f"Project {name} not found")


main_cmd = Cmd(
    "projects p|roject",
    None,
    "manage projects",
    [
        Cmd(
            "list ls",
            list_projects,
            "list projects",
            [
                Group(
                    Arg("--csv", action="store_true", help="print as CSV"),
                    Arg("--json", action="store_true", help="print as JSON"),
                    Arg("--yaml", action="store_true", help="print as YAML"),
                ),
            ],
            is_default=True,
        ),
        Cmd(
            "create",
            create_project,
            "create a project",
            [
                Arg(
                    "name",
                    help="The name of the project. Must begin with a letter, but may "
                    "contain letters, numbers, and hyphen",
                ),
                Arg("--description", help="The description of the project"),
                Arg("--owner", help="The owner of the project"),
            ],
        ),
        Cmd(
            "update",
            update_project,
            "modify a project",
            [
                Arg("projectname", help="The name of the project"),
                Arg(
                    "--name",
                    help="The new name of the project. Must begin with a letter, but may "
                    "contain letters, numbers, and hyphen",
                ),
                Arg("--description", help="The new description of the project"),
                Arg("--owner", help="The owner of the project"),
            ],
        ),
        Cmd(
            "delete",
            delete_project,
            "delete a project",
            [
                Arg("name", help="The name of the project"),
            ],
        ),
        Cmd(
            "show",
            show_project,
            "show a project",
            [
                Arg("name", help="The name of the project"),
                Group(
                    Arg("--yaml", action="store_true", help="print as YAML", default=True),
                    Arg("--json", action="store_true", help="print as JSON"),
                ),
            ],
        ),
    ],
)

args_description = [main_cmd]  # type: List[Any]



# coding: utf-8

"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictBool, StrictStr
from pydantic import Field
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class TrainedModelRegistryRequest(BaseModel):
    """
    Provides the metadata that describes how to access a named model registry to enable download of a trained model for deployment.
    """ # noqa: E501
    access_key: Optional[StrictStr] = Field(default=None, description="The access key, username or team name for the registry. * `s3` - The access key/username. * `ngc` - The optional NGC team name. * `pfs` - The optional repository name.", alias="accessKey")
    bucket: Optional[StrictStr] = Field(default=None, description="The bucket or organization name.  * `s3` - The required S3 bucket name. * `ngc` - The required NGC org name. * `pfs` - The required project name.")
    endpoint_url: Optional[StrictStr] = Field(default=None, description="The registry endpoint (host): * `s3` - The S3 registry endpoint. Required. * `openllm` - The huggingface.co compatible endpoint (default https://huggingface.co). * `ngc` - The NVIDIA NGC-compatible api endpoint (default https://api.ngc.nvidia.com). * `pfs` - The required HPE MLDM cluster (pachd) address.", alias="endpointUrl")
    insecure_https: Optional[StrictBool] = Field(default=None, description="For https endpoints, the server certificate will be accepted without validation.", alias="insecureHttps")
    modified_at: Optional[StrictStr] = Field(default=None, description="Date-time of last modification of the registry. This is a read-only field and is automatically updated.", alias="modifiedAt")
    name: StrictStr = Field(description="The name of the registry.  Must begin with a letter, but may contain letters, numbers, and hyphen.")
    secret_key: StrictStr = Field(description="The secret key is the password, secret key, or access token for the registry.  * `s3` - The secret key for the S3 bucket. * `openllm` -  The access token for huggingface.co and is supplied to the launched container via the `HF_TOKEN` environment variable. * `ngc` - The required NVIDIA NGC apikey. * `pfs` - The required authentication token (e.g. c91ab42a10a...).", alias="secretKey")
    type: StrictStr = Field(description="The type of this model registry. Must be one of the values: (s3, http, openllm, ngc, pfs). * `s3` - Configuration to enable direct model download from an s3 bucket. * `openllm` - Configuration to enable direct model download of OpenLLM models from huggingface.co. * `ngc` -  Configuration to enable direct model download from the NVIDIA NGC: AI Development Catalog. * `pfs` -  Configuration to enable direct model download from an HPE MLDM Pachyderm File System.")
    __properties: ClassVar[List[str]] = ["accessKey", "bucket", "endpointUrl", "insecureHttps", "modifiedAt", "name", "secretKey", "type"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of TrainedModelRegistryRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of TrainedModelRegistryRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "accessKey": obj.get("accessKey"),
            "bucket": obj.get("bucket"),
            "endpointUrl": obj.get("endpointUrl"),
            "insecureHttps": obj.get("insecureHttps"),
            "modifiedAt": obj.get("modifiedAt"),
            "name": obj.get("name"),
            "secretKey": obj.get("secretKey"),
            "type": obj.get("type")
        })
        return _obj


